/* Copyright 2008-2017 Research Foundation State University of New York   */

/* This file is part of QUB Online.                                       */

/* QUB Online is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation, either version 3 of the License, or      */
/* (at your option) any later version.                                    */

/* QUB Online is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License,     */
/* named LICENSE.txt, in the QUB Online program directory.  If not, see  */
/* <http://www.gnu.org/licenses/>.                                        */

/*
begin_html
<b><a href="index.html">Up</a></b>
<ul>
<li><a href="qubfast.h.html">qubfast.h</a></li>
</ul>
end_html
*/

#ifdef _WIN32
  #include <windows.h>
  #include <time.h>
  #define sleep Sleep
#else
  #include <stdlib.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
#include <math.h>
#include <string.h>

#include "qub_eventmaker.h"

int FindKnownAmp(vector<double>& known, double amp, double delta, double& ampOut)
{
  // assert(known.size() > 0);
	double minDist = fabs(amp - known[0]);
	int iMin = 0;
	for ( int i=((int) known.size())-1; i>0; --i ) {
		double dist = fabs(amp - known[i]);
		if ( dist < minDist ) {
			minDist = dist;
			iMin = i;
		}
	}
	ampOut = known[iMin];
	return (minDist < delta);
}


class ConsolidInterval
{
public:
	double interval;
	double from, to;

	ConsolidInterval(double i_from, double i_to)
		: interval(i_to - i_from), from(i_from), to(i_to)
	{}
	bool operator<(const ConsolidInterval& other) const
	{
		return (interval > other.interval); // to make the "biggest" in the heap be the smallest interval
	}
};

class Consolidator
{
public:
	vector<int>& clss;
	vector<int>& counts;
	map<double, int>& clsOfAmp;
	vector<double>& ampOfCls;
	vector<double>& known;
	int& StopFlag;
	double delta;
	typedef pair<int, vector<int> > SortAmpRec;
	typedef map<double, SortAmpRec> SortAmps;
	SortAmps sortAmps; // map from amp value to (point count, class indices) it represents
	SortAmps knownAmps;
	typedef QUB_Heap<ConsolidInterval> SortInter;
	SortInter sortInter; // heap of intervals between amps, smallest on top

	Consolidator(vector<int>& i_clss, vector<int>& i_counts, map<double, int>& i_clsOfAmp, vector<double>& i_ampOfCls, vector<double>& i_known, int& i_StopFlag);

	// this will leave you with a smaller number of classes, in order, separated by at least delta,
	// and it will leave some gaps (clss -1, count 0) in your event record as it joins neighboring events.
	void consolidate(double delta_);
	void consolidateAmps();
	void renumberClasses();
	void consolidateEvents();
};

Consolidator::Consolidator(vector<int>& i_clss, vector<int>& i_counts, map<double, int>& i_clsOfAmp, vector<double>& i_ampOfCls, vector<double>& i_known, int& i_StopFlag)
  : clss(i_clss), counts(i_counts), clsOfAmp(i_clsOfAmp), ampOfCls(i_ampOfCls), known(i_known), StopFlag(i_StopFlag)
{
  vector<int> countInCls(ampOfCls.size(), 0);
  for ( int i=((int) clss.size())-1; i>=0; --i )
    countInCls[clss[i]] += counts[i];
  for (int i=0; i<(int)ampOfCls.size(); ++i ) {
    if ( find(known.begin(), known.end(), ampOfCls[i]) != known.end() ) {
      SortAmpRec& rec = knownAmps[ampOfCls[i]];
      rec.first = countInCls[i];
      rec.second.push_back(i);
    }
    else {
      SortAmpRec& rec = sortAmps[ampOfCls[i]];
      rec.first += countInCls[i];
      rec.second.push_back(i);
    }
  }
  SortAmps::iterator ai_last = sortAmps.begin();
  SortAmps::iterator ai = ai_last;
  if ( ai != sortAmps.end() ) {
    for ( ++ai; ai != sortAmps.end(); ai_last = ai, ++ai ) {
      sortInter.push(ConsolidInterval(ai_last->first, ai->first));
    }
  }
}

// this will leave you with a smaller number of classes, in order, separated by at least delta,
// and it will leave some gaps (clss -1, count 0) in your event record as it joins neighboring events.
void Consolidator::consolidate(double delta_)
{
  delta = delta_;
  consolidateAmps();
  renumberClasses();
  consolidateEvents();
}

void Consolidator::consolidateAmps() {
  if (sortInter.size() == 0)
    return;
  while (sortInter.size() > 1 && sortInter.first().interval < delta) {
    if ( StopFlag )
      break;
    
    // we will replace the endpoints (classes) of this too-small interval with one averaged class
    
    // remove endpoints
    double from = sortInter.first().from;
    double to = sortInter.first().to;
    SortAmpRec fromRec = sortAmps[from];
    SortAmpRec toRec = sortAmps[to];
    sortAmps.erase(from);
    sortAmps.erase(to);
    
    // add endpoint from weighted average of old endpoints, sum point count, union class indices
    double newAmp = (from * fromRec.first + to * toRec.first) / (fromRec.first + toRec.first);
    SortAmpRec& newAmpRec = sortAmps[newAmp];
    newAmpRec.first = fromRec.first + toRec.first;
    for ( vector<int>::iterator ii = fromRec.second.begin(); ii != fromRec.second.end(); ++ii )
      newAmpRec.second.push_back(*ii);
    for ( vector<int>::iterator ii = toRec.second.begin(); ii != toRec.second.end(); ++ii )
      newAmpRec.second.push_back(*ii);
    
    // remove interval and neighbors
    // aka remove intervals incident on from and/or to
    // or: pop, remove the interval from "to" and the one to "from"
    sortInter.pop();
    for ( int i=0; i<2; ++i ) { // get one neighboring interval then the other
      for ( SortInter::iterator ii = sortInter.begin(); ii != sortInter.end(); ++ii ) {
	if ( (ii->from == to) || (ii->to == from) ) {
	  sortInter.erase(ii);
	  break;
	}
      }
    }
    
    // add intervals incident on new endpoint
    SortAmps::iterator otherAmp = sortAmps.find(newAmp);
    if ( otherAmp != sortAmps.begin() ) {
      --otherAmp;
      sortInter.push(ConsolidInterval(otherAmp->first, newAmp));
    }
    otherAmp = sortAmps.find(newAmp);
    ++otherAmp;
    if ( otherAmp != sortAmps.end() ) {
      sortInter.push(ConsolidInterval(newAmp, otherAmp->first));
    }
  }
}

void Consolidator::renumberClasses()
{
  clsOfAmp.clear();
  ampOfCls.clear();
  map<int, int> oldToNewCls;
  int i = 0;
  for ( SortAmps::iterator ai = knownAmps.begin(); ai != knownAmps.end(); ++i, ++ai ) {
    clsOfAmp[ai->first] = i;
    ampOfCls.push_back(ai->first);
    oldToNewCls[ai->second.second[0]] = i;
  }
  for ( SortAmps::iterator ai = sortAmps.begin(); ai != sortAmps.end(); ++i, ++ai ) {
    clsOfAmp[ai->first] = i;
    ampOfCls.push_back(ai->first);
    for ( vector<int>::iterator oci = ai->second.second.begin(); oci != ai->second.second.end(); ++oci )
      oldToNewCls[*oci] = i;
  }
  
  for ( i=((int) clss.size())-1; i>=0; --i )
    clss[i] = oldToNewCls[clss[i]];
}

void Consolidator::consolidateEvents()
{
  int last = 0;
  for ( int i=1; i<(int)clss.size(); ++i ) {
    if ( clss[i] == clss[last] ) {
      clss[i] = -1;
      counts[last] += counts[i];
      counts[i] = 0;
    }
    else {
      last = i;
    }
  }
}

// build event list by repeated add()
// if you add multiple events in a row with same amp, it joins them
// maxCount helps you break along segment bounds
// ampOfCls is DataSet:Segment:amp

// stage one: add all your events
void EventMaker::add(double amp, int count)
{
  int cls = findOrAddCls(amp);
  int ilast = ((int) clss.size()) - 1;
  if ( ilast >= 0 && cls == clss[ilast] )
    counts[ilast] += count;
  else {
    counts.push_back(count);
    clss.push_back(cls);
  }
}

// stage two: consolidate
void EventMaker::consolidate(double delta, vector<double>& known, int& StopFlag)
{
  if ( counts.size() ) {
    Consolidator cons(clss, counts, clsOfAmp, ampOfCls, known, StopFlag);
    cons.consolidate(delta);
  }
}

// stage three: pull out finished events
int EventMaker::nextEvent(int& cls, int& count, int maxCount) // returns count
{
  while ( at < (int) clss.size() && 0 == counts[at] )
    ++at;
  if ( at == ((int) clss.size()) )
    return (count = 0);
  
  cls = clss[at];
  count = counts[at];
  if ( count > maxCount ) {
    counts[at] = count - maxCount;
    count = maxCount;
  }
  else {
    ++at;
  }
  return count;
}

int EventMaker::findOrAddCls(double amp)
{
  map<double, int>::iterator ii = clsOfAmp.find(amp);
  int cls;
  if ( ii != clsOfAmp.end() ) {
    cls = ii->second;
    return cls;
  }
  // (else):
  cls = (int) ampOfCls.size();
  clsOfAmp[amp] = cls;
  ampOfCls.push_back( amp );
  return cls;
}



class QUB_EventMaker_ByGauss_Sort_Mean {
public:
  vector<IncrementalRSS>& dist;
  QUB_EventMaker_ByGauss_Sort_Mean(vector<IncrementalRSS>& _dist)
    : dist(_dist)
  {}
  bool operator()(int a, int b) {
    return (dist[a].mean < dist[b].mean);
  }
};

class QUB_EventMaker_ByGauss_Sort_Delta {
public:
  vector<double>& deltas;
  QUB_EventMaker_ByGauss_Sort_Delta(vector<double>& _deltas)
    : deltas(_deltas)
  {}
  bool operator()(int a, int b) {
    return (deltas[a] < deltas[b]);
  }
};

class QUB_EventMaker_ByGauss_Sort_Size_Dec {
public:
  vector<IncrementalRSS>& dist;
  vector<int>& rewrite;
  QUB_EventMaker_ByGauss_Sort_Size_Dec(vector<IncrementalRSS>& _dist, vector<int>& _rewrite)
    : dist(_dist), rewrite(_rewrite)
  {}
  bool operator()(int a, int b) {
    int Na = (rewrite[a] == a) ? (int) dist[a].n : 0;
    int Nb = (rewrite[b] == b) ? (int) dist[b].n : 0;
    return (Na > Nb);
  }
};



int QUB_EventMaker_ByGauss::get_mean_index(int c)
{
  vector<int>::iterator mm = lower_bound(mean_order.begin(), mean_order.end(), c, QUB_EventMaker_ByGauss_Sort_Mean(dist));
  if ( mm != mean_order.end() )
    return (int) (mm - mean_order.begin());
  return -1;
}

void QUB_EventMaker_ByGauss::combine(int c, int d)
{ // the combined class has index c
  dist[c] += dist[d];
  mean_order.erase(lower_bound(mean_order.begin(), mean_order.end(), d, QUB_EventMaker_ByGauss_Sort_Mean(dist)));
  sort(mean_order.begin(), mean_order.end(), QUB_EventMaker_ByGauss_Sort_Mean(dist));
  byMean[dist[c].mean] = c;
  int N = (int) rewrite.size();
  for (int i=0; i<N; ++i)
    if ( rewrite[i] == d )
      rewrite[i] = c;
}

bool QUB_EventMaker_ByGauss::combine_if_within_std(int c, int d, double delta)
{ // use sample_std()?
  if ( (dist[c] + dist[d]).sample_std() <= delta ) {
    combine(c, d);
    return true;
  }
  return false;
}

bool QUB_EventMaker_ByGauss::combine_neighbor(int c, double delta)
{
  int N = (int) mean_order.size();
  vector<double> deltas;
  vector<int> delta_cls, delta_ix;
  int mc = get_mean_index(c);
  // get (delta, index) of classes with mean in (c.mean +/- delta)
  for (int sgn=-1; sgn<2; sgn+=2) { // sgn in [-1, +1]
    int md = mc + sgn;
    while ( md >= 0 && md < N ) {
      if ( exists(mean_order[md]) ) {
	double d = fabs(dist[c].mean - dist[mean_order[md]].mean);
	if ( d > delta )
	  break;
	deltas.push_back(d);
	delta_cls.push_back(mean_order[md]);
	delta_ix.push_back((int) delta_ix.size());
      }
      md += sgn;
    }
  }
  sort(delta_ix.begin(), delta_ix.end(), QUB_EventMaker_ByGauss_Sort_Delta(deltas));
  for (vector<int>::iterator dd=delta_ix.begin(); dd!=delta_ix.end(); ++dd)
    if ( combine_if_within_std(c, delta_cls[*dd], delta) )
      return true;
  return false;
}

int QUB_EventMaker_ByGauss::indices_by_size_dec(vector<int>& indices)
{
  int N = (int) dist.size();
  indices.resize(N);
  for (int i=0; i<N; ++i)
    indices[i] = i;
  sort(indices.begin(), indices.end(), QUB_EventMaker_ByGauss_Sort_Size_Dec(dist, rewrite));
  return N;
}

void QUB_EventMaker_ByGauss::renumber(int *cc, int n)
{
  for (int i=0; i<n; ++i)
    cc[i] = rewrite[cc[i]];
}

void QUB_EventMaker_ByGauss::renumber_classes()
{
  int Nc = (int) dist.size();
  int i_compact = 0;
  vector<int> collapsed(Nc);
  for (int c=0; c<Nc; ++c) {
    if ( exists(c) ) {
      collapsed[c] = i_compact;
      dist[i_compact] = dist[c];
      ++i_compact;
    }
  }
  for (int c=0; c<Nc; ++c)
    rewrite[c] = collapsed[rewrite[c]];
  dist.resize(i_compact);

  collapsed.resize(i_compact);
  for (int c=0; c<i_compact; ++c)
    collapsed[c] = c;
  sort(collapsed.begin(), collapsed.end(), QUB_EventMaker_ByGauss_Sort_Mean(dist));
  vector<int> sorted(i_compact);
  for (int c=0; c<i_compact; ++c)
    sorted[collapsed[c]] = c; // inverse of collapsed
  vector<IncrementalRSS> last_dist = dist;
  vector<int> last_rewrite = rewrite;
  for (int c=0; c<i_compact; ++c)
    dist[c] = last_dist[collapsed[c]];
  for (int c=0; c<Nc; ++c)
    rewrite[c] = sorted[last_rewrite[c]];

  renumber(& clss[0], (int) clss.size());
}

int QUB_EventMaker_ByGauss::nextEvent(int& cls, int& count, int maxCount) // returns count
{
  while ( at < (int) clss.size() && 0 == counts[at] )
    ++at;
  if ( at == ((int) clss.size()) )
    return (count = 0);
  
  cls = clss[at];
  count = counts[at];
  if ( count > maxCount ) {
    counts[at] = count - maxCount;
    count = maxCount;
  }
  else {
    ++at;
  }
  return count;
}


int QUB_EventMaker_ByGauss::add_event(double mean, double rss, int n)
{
  int c;
  map<double, int>::iterator existing = byMean.find(mean);
  if ( existing != byMean.end() ) {
    c = existing->second;
    dist[c].rss += rss;
    dist[c].n += n;
  } else {
    c = (int) dist.size();
    dist.resize(c+1);
    rewrite.resize(c+1);
    rewrite[c] = c;
    dist[c].mean = mean;
    dist[c].rss = rss;
    dist[c].n = n;
    byMean[mean] = c;
  }
  int ilast = ((int) clss.size()) - 1;
  if ( ilast >= 0 && c == clss[ilast] )
    counts[ilast] += n;
  else {
    counts.push_back(n);
    clss.push_back(c);
  }
  return c;
}

void QUB_EventMaker_ByGauss::done_add()
{
  mean_order.resize(dist.size());
  for (int c=(int)dist.size()-1; c>=0; --c)
    mean_order[c] = c;
  sort(mean_order.begin(), mean_order.end(), QUB_EventMaker_ByGauss_Sort_Mean(dist));
}

void QUB_EventMaker_ByGauss::combine_all(double delta)
{
  vector<int> byN;
  indices_by_size_dec(byN);
  for (vector<int>::iterator c=byN.begin(); c!=byN.end(); ++c) {
    if ( exists(*c) ) {
      while ( combine_neighbor(*c, delta) )
	{}
    }
  }
  renumber_classes();
}

int QUB_EventMaker_ByGauss::get_next_dwells(int offset_in_data, int sample_count, int *firsts, int *lasts, int *classes)
{
  int last = offset_in_data - 1;
  int ndwt = 0;
  int cls, count;
  while ( sample_count ) {
    nextEvent(cls, count, sample_count);
    if ( ! count )
      break; // should never happen if add's counts match get_next_dwells's
    sample_count -= count;
    firsts[ndwt] = last + 1;
    last += count;
    lasts[ndwt] = last;
    classes[ndwt] = cls;
    ++ndwt;
  }
  return ndwt;
}

int QUB_EventMaker_ByGauss::class_count()
{
  return (int) dist.size();
}

void QUB_EventMaker_ByGauss::get_dist(double *mean, double *std)
{
  int N = (int) dist.size();
  for (int c=0; c<N; ++c) {
    *(mean++) = dist[c].mean;
    *(std++) = dist[c].sample_std();
  }
}


      
int delete_short_events(int ndwt, int *firsts, int *lasts, int *classes, int min_nsamp)
{
  int i = 0;
  int init_shorts = 0;
  int oi = 0;
  while ( i<ndwt && (lasts[i] - firsts[i] + 1) < min_nsamp ) {
    init_shorts += (lasts[i] - firsts[i] + 1);
    ++i;
  }
  if ( i == ndwt ) {
    lasts[0] = lasts[ndwt-1];
    classes[0] = -1;
    return 0;
  }
  while ( i<ndwt ) {
    if ( oi && (classes[i] == classes[oi-1]) ) // join same-class
      lasts[oi-1] = lasts[i];
    else if ( oi && ((lasts[i] - firsts[i] + 1) < min_nsamp) ) // join too-short
      lasts[oi-1] = lasts[i];
    else {
      classes[oi] = classes[i];
      firsts[oi] = firsts[i] - init_shorts;
      lasts[oi] = lasts[i];
      ++oi;
      init_shorts = 0;
    }
    ++i;
  }
  return oi;
}

int delete_unused_classes(int ndwt, int *classes, int nc, double *mean, double *std)
{
  vector<int> seen(nc);
  for (int i=0; i<ndwt; ++i)
    seen[classes[i]] = 1;
  int c = 0;
  for (int i=0; i<nc; ++i) {
    if ( seen[i] ) {
      seen[i] = c;
      if ( c < i ) {
	mean[c] = mean[i];
	std[c] = std[i];
      }
      ++c;
    }
  }
  for (int i=0; i<ndwt; ++i)
    classes[i] = seen[classes[i]];
  return c;
}

